#Exercise 3 template for CHE210D

#import python modules
import numpy as np
import pylab as py
import time
import cPickle as pickle
import cProfile

from mpl_toolkits.mplot3d import Axes3D
from mayavi import mlab

#import compiled fortran library
import ex4lib
import atomwrite

import sys
import os

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis

#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
np.random.seed(321)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global EMat, kx, M, k, L, Cut, Types, kx
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [ex4lib.calcenergy(Pos + NormDir * x, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = ex4lib.calcenergy(Pos + NormDir * dx, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [ex4lib.calcenergy(Pos + NormDir * 2. * dx, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut)]
        
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [ex4lib.calcenergy(Pos + NormDir * Dists[-1], Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = ex4lib.calcenergy(Pos + NormDir * d3, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = ex4lib.calcenergy(Pos + NormDir * d3, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin,Force=True)
        else:
            #cs = [[0,10*i,0] for i in range(N)]
            #cs[1] = [150,0,0]
            #cs[2] = [0,0,150]
            #cs[3] = [100,0,150]
            #initialize the visualization window
            atomvis.Init(PosMin)#,AtomColor=cs)
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global EMat, kx, M, k, L, Cut, Types
    
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = ex4lib.calcenergyforces(Pos, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = ex4lib.calcenergyforces(Pos, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)    
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    
    Nyz = int((N/2)**(1.0/3.0)) + 1
    Nx = 2 * Nyz
    
    LatSpac = L / Nyz
    
    ryz = LatSpac * np.arange(Nyz) - 0.5*L
    rx = LatSpac * np.arange(Nx) - L + 1.0
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in rx:
        for y in ryz:
            for z in ryz:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                
                if np.abs(-L - Pos[i][0]) < 1.0:
                    Pos[i][0] = -L + 1.0
                if np.abs(L - Pos[i][0]) < 1.0:
                    Pos[i][0] = L - 1.0
                
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global EMat, kx, M, k, L, Cut, Types
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = ex4lib.calcenergyforces(Pos, Types, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))
    
#-----------------Additional Functions--------------------
def TakeMDStep(Pos, Types, Vel, Accel, EMat, kx, dx, M, k, L, Cut, dt):
    return ex4lib.vvintegrate(Pos, Types, Vel, Accel, EMat, kx, dx, M, k, L, Cut, dt)
    

def RunMD(NStepsProd):
    
    global EMat, kx, M, NA, k, L, Cut, N, T, Types, PdbWrite, Corners, PosRecord, NChains
    
    NStepsEquil1 = 10000
    NStepsEquil2 = 10000
    RecordFrequency = 5000

    RescaleFrequency = 250
    
    E1 = []
    E = np.zeros(NStepsEquil2)
 
    dt = 0.01
        
    Pos, Vel, Accel = InitializeMD()

    #First equilibration step
    for i in range(NStepsEquil1):
        #Take an MD step
        Pos, Vel, Accel, KEnergy, PEnergy = TakeMDStep(Pos, Types, Vel, Accel, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut, dt)
        
        if i % RescaleFrequency == 0:
            Vel = RescaleVelocities(Vel,T)
            E1.append(PEnergy + KEnergy)
    
    py.plot(E1)
    py.title("E(t) vs t for Equilibration Step 1")
    py.savefig("E equilibrium 1.png")
    py.clf()
            
    #Second equilibration step
    for i in range(NStepsEquil2):
        if i % RescaleFrequency == 0:
            Vel = RescaleVelocities(Vel,T)
        #Take an MD step
        Pos, Vel, Accel, KEnergy, PEnergy = TakeMDStep(Pos, Types, Vel, Accel, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut, dt)
        E[i] = KEnergy + PEnergy
        
    TargetE = np.mean(E)
    #Calculate lambda required to make KE = <E> - PE
    lam = np.sqrt(2.*(TargetE-PEnergy)/(np.sum(Vel*Vel)))
    #Now rescale current velocities with the calculated lambda
    Vel = Vel * lam
        
    #Production run
    for i in range(NStepsProd):
        if i%(NStepsProd/10) == 0:
            print "Step = {}/10".format((10*i)/NStepsProd)
        if i % RecordFrequency == 0:
            #UpdateZProfile(Pos)
            #CalculateRGs(Pos)
            #UpdateCorrelationFunction(Pos)
            Pdb.write(np.array(list(Pos)+Corners))
            PosRecord.append(Pos[0:NA].copy())
            PEnergyRecord.append(PEnergy)
        Pos, Vel, Accel, KEnergy, PEnergy = TakeMDStep(Pos, Types, Vel, Accel, EMat, kx, GetDisplacementFromTarget(Pos), M, k, L, Cut, dt)

def CalculateRGs(Pos):
    global Rg, NChains, M
    
    avg = []
    
    for i in range(NChains):
        
        dr = np.sqrt(np.sum(np.square(Pos[i+M-1] - Pos[i])))
        avg.append(dr)
    
    Rg.append(np.mean(avg))
    
    
def UpdateZProfile(Pos):
    global RhoZ, Types, L

    BinWidth = 2*L
    NBins = 12
    BinWidth = BinWidth / NBins

    bins = np.zeros([np.max(Types)+1, NBins])
    
    for i in range(N):
        z = int((Pos[i][0]+L) / BinWidth)
        bins[Types[i]][z] += 1
    
    RhoZ.append(bins / (2*L**3 / NBins))
    
def UpdateCorrelationFunction(Pos):
    global CorrelationBins, Types
    
    BinWidth = np.array([2*L, L, L])
    NBins = np.array([20,10,10])
    BinWidth = BinWidth / NBins
    
    CurrentCorrelation = np.zeros([20,10,10])
    CorrelationBins.append(CurrentCorrelation)
    
    for i in range(N):
        for j in range(i,N):
            if i!=j and Types[i] == Types[j] and Types[i] == 0 :
                dr = np.abs(Pos[j] - Pos[i])
                dr[1:3] = dr[1:3] - L * np.round_(dr[1:3] / L)
                #Shift the zero point so dr = {0,L} instead of {-L/2,L/2}
                #dr = dr + np.array([L, L/2.0, L/2.0])

                #Locate the appropriate bin for this dr
                binz = np.array(dr / BinWidth, int)
                CurrentCorrelation[binz[0]][binz[1]][binz[2]] += 1
                
                
    
def CalculateDiffusivityMD(NStepsProd):
    global L, Cut, N, M, T
    
    NStepsEquil1 = 10000
    NStepsEquil2 = 10000
    RecordFrequency = 250

    RescaleFrequency = 250
    
    E = np.zeros(NStepsEquil2)
    t = []
    rms = []
    
    dt = 0.005
        
    Pos, Vel, Accel = InitializeMD()

    #First equilibration step
    for i in range(NStepsEquil1):
        if i % RescaleFrequency == 0:
            Vel = RescaleVelocities(Vel,T)
        #Take an MD step
        Pos, Vel, Accel, KEnergy, PEnergy = ex4lib.vvintegrate(Pos, Vel, Accel, M, k, L, Cut, dt)
            
    #Second equilibration step
    for i in range(NStepsEquil2):
        if i % RescaleFrequency == 0:
            Vel = RescaleVelocities(Vel,T)
        #Take an MD step
        Pos, Vel, Accel, KEnergy, PEnergy = ex4lib.vvintegrate(Pos, Vel, Accel, M, k, L, Cut, dt)
        E[i] = KEnergy + PEnergy
        
    TargetE = np.mean(E)
    #Calculate lambda required to make KE = <E> - PE
    lam = np.sqrt(2.*(TargetE-PEnergy)/(np.sum(Vel*Vel)))
    #Now rescale current velocities with the calculated lamda
    Vel = Vel * lam

    #Copy our starting configuration as a reference for RMS displacement calculations later    
    Pos0 = Pos.copy()
    
    DiffusivityList = []
    
    #Production run
    for i in range(NStepsProd):
            #if using visualization, update the display
        if UseVisual:
            if atomvis.Initialized:
                #update the positions
                atomvis.Update(Pos)
            else:
                #initialize the visualization window
                atomvis.Init(Pos)

        if i % RecordFrequency == 0:
            DiffusivityAnalysis(DiffusivityList, [Pos, i])
            #Record the time
            t.append(dt*i)
            #dr
            dr = np.array(Pos - Pos0)
            #List of displacement vectors |dr_i|^2
            dr = np.sum(dr * dr,axis=1)
            #Average <|dr_i|^2> over all particles
            rms.append(np.mean(dr))
        #Take MD Step
        Pos, Vel, Accel, KEnergy, PEnergy = ex4lib.vvintegrate(Pos, Vel, Accel, M, k, L, Cut, dt)
    
    f = file("data/part-e/MD-M={}".format(M),'w')
    pickle.dump(DiffusivityList[1],f)
    f.close()
    #slope, intercept = np.polyfit(t[len(t)/2:],rms[len(rms)/2:],1)
    #return slope / 6., rms

def InitializeMD():
    global L, Cut, proposed, accepted, N, M, T, Pdb, Corners
        
    #Start MC Simulation with new random conditions
    np.random.seed(int(os.times()[4]*100))

    Pos = InitPositions(N,L)
    
    PE, Pos = ConjugateGradient(Pos, .001, 1.e-8, 1.e-10)

    if UseVisual:
        atomvis.Update(Pos,Force=True)
    Accel = InitAccel(Pos)
    Vel = InitVelocities(N, T)
    
    return Pos, Accel, Vel
        
def ShowPositions(Pos):
    global L
    fig = py.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.set_xlabel('X Label')
    ax.set_ylabel('Y Label')
    ax.set_zlabel('Z Label')
    x,y,z=np.transpose(Pos-L*np.array(Pos/L,int))
    ax.scatter(x,y,z)    
    ax.set_xlim3d(-L/2.0,L/2.0)
    ax.set_ylim3d(-L/2.0,L/2.0)
    ax.set_zlim3d(-L/2.0,L/2.0)

    py.show()

def InitSimulation():
    global EMat, E, Eab, M, NChains, NA, Types, Names, rho, N, L, Corners, T, PosRecord, PEnergyRecord
        
    EMat = [[E, Eab],[Eab,E]]

    CopolymerNames = ['P']*(NA) + ['N']*(M-NA)
    CopolymerTypes = [0]*M

    Types = CopolymerTypes * NChains + [1]*(N-M)
    Names = CopolymerNames * NChains + ['O']*(N-M) + ['C']*8
    
    PosRecord = []
    PEnergyRecord = []
    
    p = [2,-2,0,-N/rho]
    L = np.float(np.roots(p)[0])
    
    print "Simulation Length L={}".format(L)
    
    Corners = [ [-L,-L/2.0,-L/2.0],
                [L,-L/2.0,-L/2.0],
                [L,-L/2.0,L/2.0],
                [-L,-L/2.0,L/2.0],
                
                [-L,L/2.0,-L/2.0],
                [L,L/2.0,-L/2.0],
                [L,L/2.0,L/2.0],
                [-L,L/2.0,L/2.0],
    ]   

def GetDisplacementFromTarget(Pos):
    global M, xtarget, NA
    r = np.array((Pos[0:NA]),float)
    dx = np.mean(xtarget-r[:,0])
    return dx    


#==============MC Analysis Functions===========================
    
#### YOUR CODE HERE ####
kx=5

k=3000.0
E = 1
Eab = 1
EMat = []

N = 300
M = 60
NA = 60
NChains = 1

Types = []
Names = []

rho = 0.65
L = 1.0

Corners = []   

#Statistics Stuff
RhoZ = []
Rg = []
stats = [RhoZ, Rg]

PosRecord = []
PEnergyRecord = []

Stats = [PosRecord, PEnergyRecord]

#L = (N/rho)**(1./3.)
T = 1.1
Run = 0
XTargets = np.arange(-0.5,4.0,0.75)
RunSimulation = True

for rho in [0.65]:
    for T in [1.1,1.2]:
        print "Starting Run {}".format(Run)
        InitSimulation()
        for xtarget in XTargets :
            FileName = "Pos-x={}-M={}-NA={}-Run={}-T={}-rho={}".format(xtarget,M,NA,Run,T,rho)
            InitSimulation()

            Pdb = atomwrite.pdbfile("Data/Anim-x={}-M={}-NA={}-T={}-rho={}.pdb".format(xtarget,M,NA,T,rho), L, AtomNames = Names)
            RunMD(500000)
            Pdb.close()

            f = file("Data/"+FileName,'w')
            pickle.dump([PosRecord, PEnergyRecord],f)
            f.close()
